package gst;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DateFormatSymbols;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Locale;

import org.apache.commons.lang3.exception.ExceptionUtils;

import client.Message;
import client.ProcessRecord;
import client.SettingFile;
import client.Warning;
import client.Warning.Type;

/**
 * 
 * 因為在比賽進行中，有可能遭遇斷線、Search 崩潰、Client 出錯等狀況出現，比賽會被迫中斷，
 * 因不算正常結束，故無法從 Server 端收到 log 檔案，造成 User Debug 困難，也無從得知比賽狀況，
 * 因此， Client 端也會記錄比賽的 log 檔案，方便之後觀看，但只要比賽正常結束，
 * 就會刪除此檔案並儲存由 Server 端送來的 log 檔案，一切以 Server 的為準。
 * 
 * @author ldslds449
 * @see client.ProcessRecord
 */
class GST_ProcessRecord extends ProcessRecord{
	private String fileName;
	private static String dir;
	private String playerName;
	private String oppoPlayerName;
	
	private String[] board;
	private int[] remain_chess;
	
	private int timeLimit;
	
	private int gameStarted;
	private String start_time;
	
	private ArrayList<String> moves = new ArrayList<String>();
	private ArrayList<Long> times = new ArrayList<Long>();
	private ArrayList<Integer> eats = new ArrayList<Integer>();
	
	private String searchInfo;
	
	/**
	 * 
	 * constructor。紀錄和處理 log 檔案標頭參數。
	 * 
	 * @param round 這場比賽是第幾回合的比賽。
	 * @param midBoard 比賽開始之前的初始盤面。
	 * @param setting 這房間的房間參數。
	 */
	GST_ProcessRecord(int round, String midBoard, String searchInfo, SettingFile setting){
		this.fileName = String.format("%d temp.txt", round);
		this.timeLimit = ((GST_SettingFile)setting).timeLimit;
		this.searchInfo = searchInfo;
		
		String[] temp = midBoard.split(",");
		// record board
		this.board = new String[36];
//		final char kind[] = {'-', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'X', 'a', 'b', 'c', 'd', 'e', 'f','g' ,'h'};
		final char kind[] = {'-', 'R', 'B', 'X', 'r', 'b'};
		for(int i = 0; i < 36; i++){
			this.board[i] = String.valueOf(kind[Integer.parseInt(temp[i])]);
		}
		
		// record remain chess
		this.remain_chess = new int[4];
		for(int i = 0; i < 4; i++){
			this.remain_chess[i] = Integer.parseInt(temp[36 + i]);
		}
		
		// record gameStarted
		gameStarted = Integer.parseInt(temp[40]);
		// record step
		int step_size = Integer.parseInt(temp[41]);
		for(int i = 0; i < step_size; i++){
			moves.add(temp[42 + i]);
		}
		// record eats
		int eats_size = Integer.parseInt(temp[42 + step_size]);
		for(int i = 0; i < eats_size; i++){
			eats.add(Integer.parseInt(temp[43 + step_size + i]));
		}
		// record time
		int time_size = Integer.parseInt(temp[43 + step_size + eats_size]);
		for(int i = 0; i < time_size; i++){
			times.add(Long.parseLong(temp[44 + step_size + eats_size + i]));
		}
//		this.board = new String[32];
//		final char kind[] = {'-', 'K', 'G', 'M', 'R', 'N', 'C', 'P', 'X', 'k', 'g', 'm', 'r', 'n', 'c', 'p'};
//		for(int i = 0; i < 32; i++){
//			this.board[i] = String.valueOf(kind[Integer.parseInt(temp[i])]);
//		}
//		
//		// record remain chess
//		this.remain_chess = new int[14];
//		for(int i = 0; i < 14; i++){
//			this.remain_chess[i] = Integer.parseInt(temp[32 + i]);
//		}
//		
//		// record color
//		gameStarted = Integer.parseInt(temp[46]);
//		// record step
//		int step_size = Integer.parseInt(temp[47]);
//		for(int i = 0; i < step_size; i++){
//			moves.add(temp[48 + i]);
//		}
//		// record time
//		int time_size = Integer.parseInt(temp[47 + step_size + 1]);
//		for(int i = 0; i < time_size; i++){
//			times.add(Long.parseLong(temp[48 + step_size + 1 + i]));
//		}
		
		// record start time
		Date now = new Date();
		SimpleDateFormat format = new SimpleDateFormat("K:ma MMM d, y", Locale.ENGLISH);
		DateFormatSymbols symbols = new DateFormatSymbols(Locale.ENGLISH);
        // override symbols
        symbols.setAmPmStrings(new String[] { "am", "pm" });
        format.setDateFormatSymbols(symbols);
        start_time = format.format(now);
	}
	
	/**
	 * 刪除此由 Client 自行紀錄的 log 檔案。
	 */
	synchronized public void deleteTempFile(){
		new File(String.format("%s%s%s", dir, File.separator, fileName)).delete();
	}
	
	/**
	 * 
	 * 設置玩家雙方名稱，因為紀錄 log 檔案需要。
	 * 
	 * @param playerName 我方玩家名稱。
	 * @param oppoPlayerName 對方玩家名稱。
	 */
	void setPlayerName(String playerName, String oppoPlayerName){
		this.playerName = playerName;
		this.oppoPlayerName = oppoPlayerName;
	}
	
	/**
	 * 
	 * 紀錄走步和時間。
	 * 
	 * @param data 走步和時間(就是 Server 傳送過來的完整資料)。
	 * @param isFirst 是否為第一步。
	 * @throws Warning 發生錯誤。
	 */
	void addData(String data, boolean isFirst) throws Warning{
		// record moves
		String[] move = data.split(",");
		moves.add(move[0] + "," + move[1]);
		// record times
		if(isFirst){
			times.add(Long.parseLong(move[6]));
			times.add(Long.parseLong(move[7]));
		}else{
			times.add(Long.parseLong(move[7]));
		}
	}
	
	/**
	 * 紀錄玩家執子顏色(log 檔案需要用到)。
	 * @param color 玩家顏色。
	 */
	void setColor(int color){
		this.gameStarted = color;
	}
	
	/**
	 * 輸出遊戲資料到 log 檔案(不是 append 的，是全部從頭輸出的，因為要 insert 太困難了。)
	 * @param stdout_stamp standard output 的走步輸出行數 stamp(用來讓 web 知道這走步當時大約的 log 行數是第幾行，方便 user 查看)。
	 * @param stderr_stamp standard error 的走步輸出行數 stamp(用來讓 web 知道這走步當時大約的 log 行數是第幾行，方便 user 查看)。
	 * @param stdout Search 的 standard output。
	 * @param stderr Search 的 standard error。
	 * @throws Warning 發生錯誤。
	 */
	synchronized void writeTempFile(String stdout_stamp, String stderr_stamp, String stdout, String stderr) throws Warning{
		
		try {
			StringBuilder sb = new StringBuilder();
			BufferedWriter BW = new BufferedWriter(new FileWriter(String.format("%s%s%s", dir, File.separator, fileName)));
			
			// output warning
			sb.append("This is TEMPORARILY file. You can use it for DEBUG.");
			sb.append(System.getProperty("line.separator"));
			// output time
			sb.append(String.format("* %s", start_time));
			sb.append(System.getProperty("line.separator"));
			// output game info
			sb.append(String.format("* %s vs %s", playerName, oppoPlayerName));
			sb.append(System.getProperty("line.separator"));
			// output remain piece
			sb.append('*');
			for(int i = 0; i < 4; i++){
				sb.append(' ');
				sb.append(this.remain_chess[i]);
			}
			sb.append(System.getProperty("line.separator"));
			// output initial board
			for(int i = 0; i < 6; i++){
				sb.append('*');
				for(int j = 0; j < 6; j++){
					sb.append(' ');
					sb.append(this.board[i * 6 + j]);
				}
				sb.append(System.getProperty("line.separator"));
			}
			// output time limit
			sb.append(String.format("* time %d", timeLimit));
			sb.append(System.getProperty("line.separator"));
			// output history step
			for(int i = 0; i < moves.size(); i++){
				if(i % 2 == 0){
					sb.append(String.format("* %d. ", i/2 + 1));
				}
				sb.append(moves.get(i));
				if(i % 2 == 0){
					sb.append(' ');
				}
				if(i % 2 == 1 || i == moves.size() - 1){
					sb.append(System.getProperty("line.separator"));
				}
			}
			if(moves.size() > 0){
				sb.append("* Comment 0 0");
				sb.append(System.getProperty("line.separator"));
			}
			// output history time
			for(int i = 0; i < times.size(); i++){
				sb.append(String.format("%d. ", i));
				sb.append(times.get(i));
				sb.append(System.getProperty("line.separator"));
			}
			sb.append(System.getProperty("line.separator"));
			
			BW.write(sb.toString());
			
			
			// output stdout
			BW.write(stdout_stamp);
			BW.write(System.getProperty("line.separator"));
			BW.write(stdout);
			BW.write(System.getProperty("line.separator"));
			// output stderr
			BW.write(stderr_stamp);
			BW.write(System.getProperty("line.separator"));
			BW.write(stderr);
			BW.write(System.getProperty("line.separator"));
			
			BW.write("###");
			BW.write(System.getProperty("line.separator"));
			BW.write(searchInfo);
			BW.write("###");
			BW.write(System.getProperty("line.separator"));
			
			BW.flush();
			BW.close();
		} catch (IOException e) {
			Message.printError("ERROR: WRITE TEMP BOARD TXT");
			Message.printError("BOARDNAME: %s\n", fileName);
			Message.printError(ExceptionUtils.getStackTrace(e));
			
			throw new Warning(Type.WRITE_FILE_ERROR, "WRITE TEMP BOARD TXT ERROR");
		}
	}
	
	/**
	 * 設置 log 檔案的輸出資料夾。
	 * @param dir log 檔案的輸出資料夾。
	 */
	void setDir(String dir){
		GST_ProcessRecord.dir = dir;
	}
	
	/**
	 * 如果是用中盤檔案開局，會需要位移(shift)每一手走步的 time stamp。
	 * @param firstTime 第一手的時間。
	 */
	void modifyTimeStamp(long firstTime){
		long diff_time = firstTime - times.get(times.size() - 1);
		times.remove(times.size() - 1);
		
		for(int i = 0; i < times.size(); i++){
			times.set(i, times.get(i) + diff_time);
		}
	}
	
	/**
	 * 如果有玩家進行重新進房，必須補償server lock時間，所以需要位移(shift)每一手走步的 time stamp。
	 * @param lastTimeStamp 上一手走步的 time stamp。
	 */
	void shiftTimeStamp(long lastTimeStamp) {
		long diff_time = lastTimeStamp - times.get(times.size() - 1);
		if(diff_time > 0) {
			for(int i = 0; i < times.size(); i++){
				times.set(i, times.get(i) + diff_time);
			}
		}
	}
}
