package gst;

import org.apache.commons.lang3.exception.ExceptionUtils;

import client.Message;
import client.Packet;
import client.Room_SettingFile.BoardDisconnectAutoRestart;
import client.Room_SettingFile.BoardOpenType;
import client.SettingFile;
import client.SettingFile.STARTMODE;
import client.Warning;

/**
 * 
 * 處理有關人類手動模式下的連接事宜。
 * 
 * @author ldslds449
 * @see gst.SearchConnect
 */
class Human extends SearchConnect{
	
	private GST_Board board;
	private String color;
	private int playerNum;
	
	private boolean oppoIsHuamn;
	private boolean isLocal;
	private BooleanHolder isFinish = null;

	private final int TIMEOUT = 50;
	
	/**
	 * constructor，因為是拿 GST_Board 包裝成 Search 的接口模式，所以要傳入 GST_Board 所需的初始參數
	 * 另外，因為針對有些功能上有線上伺服器和本地伺服器的區別限制，所以要 isLocal 來限制某些功能。
	 * @param type 盤面開啟方式 (GST_Board 初始用)。
	 * @param startMode 開房還是加房 (GST_Board 初始用)。
	 * @param room 房間頁面 (GST_Board 初始用)。
	 * @param isLocal 判斷是否為本地伺服器。
	 * @param isCLI 是否目前為 CLI mode。
	 * @throws Warning 發生錯誤。
	 */
	Human(BoardOpenType type, STARTMODE startMode, BoardDisconnectAutoRestart restart, GST_Room room, GST_Game game, boolean isLocal, boolean isCLI, boolean isPlay, boolean needCloseWeb) throws Warning{
		
		this.board = new GST_Board(type, startMode, restart, room, game, isCLI, isPlay, needCloseWeb);
		// run board server
		this.board.start();
		// wait for Board connect
		Message.printInfo(">> Wait for Board connecting");
		synchronized (this.board) {
			try {
				this.board.wait();
			} catch (Exception e) {}
		}
		Message.printInfo(">> Board connected");
		
		this.isLocal = isLocal;
		this.color = "unknown";
	}
	
	/**
	 * {@inheritDoc}
	 * 要求 web 產生走步。
	 * @return Packet class({@link client.Packet}) 的封包，內含動作代碼和走步。
	 */
	@Override
	protected Packet askMove(String ghostCodeInfo) throws Warning {
		while(true){
			board.send(String.format("genmove %d", playerNum));
			
			// wait for data
			String result = "";
			do{
				try {
					result = board.recv(TIMEOUT);
				} catch(Warning w){
					switch(w.getType()){
					case WEBSERVER_RECV_TIMEOUT:
						if(isFinish.getValue() == true){
							return null;
						}else{
							continue;
						}
					default:
						Message.printError(ExceptionUtils.getStackTrace(w));
						return null;
					}
				}
			}while(result.equals(""));
			
			Packet p = new Packet();
			if(result.contains("MOV:")){
				p.setAction(GST_Action.Move);
				p.setData(result.split(":")[1]);
				return p;
			}else if(result.contains("resign")){
				p.setAction(GST_Action.Resign);
				return p;
			}else if(result.contains("drawOffer")){
				if(oppoIsHuamn && !isLocal){
					p.setAction(GST_Action.DrawOffer);
					return p;
				}else{
					// reject
					sendCommandResult("drawOffer", 0);
				}
			}else if(result.contains("abort")){
				if(isLocal){
					p.setAction(GST_Action.Abort);
					return p;
				}else{
					// reject
					sendCommandResult("abort", 0);
				}
			}
		}
	}
	
	@Override
	protected Packet askSetRed() throws Warning {
		while(true){
			board.send("set");
			
			// wait for data
			String result = "";
			do{
				try {
					result = board.recv(TIMEOUT);
				} catch(Warning w){
					switch(w.getType()){
					case WEBSERVER_RECV_TIMEOUT:
						if(isFinish.getValue() == true){
							return null;
						}else{
							continue;
						}
					default:
						Message.printError(ExceptionUtils.getStackTrace(w));
						return null;
					}
				}
			}while(!(result.split(":")[0].equals("SET")));
			
			if (result.endsWith("\r\n")) {
                result = result.substring(0, result.length() - 2);
            }
			
			Packet p = new Packet();
			p.setAction(GST_Action.SetRed);
			p.setData(result.split(":")[1]);
			return p;
		}
	}
	
	/**
	 * {@inheritDoc}
	 * 傳送走步給 web。
	 * @param data Server 傳送過來的走步完整資料(包含走步、時間等)。
	 * @param noEatFlip 無吃翻數字。
	 * @param step_time 當前走步所花費的時間。
	 * @param total_spent_time 此走步玩家到當前走步所花費的時間累計。
	 * @param time 我方玩家所剩餘的時間。
	 * @param oppoTime 對手剩餘時間。
	 * @param isLast 這步是否為最後一步，避免 Search 多走一步。
	 */
	@Override
	protected void sendMove(String data, int capturePieceColor, int noEatFlip, int step_time, int total_spent_time, int time, int oppoTime, boolean isLast) throws Warning {
		String[] split_data = data.split(",");
		board.sendMove(split_data[0]+" "+split_data[1], 
				capturePieceColor,
				noEatFlip, 
				step_time, 
				total_spent_time, 
				(playerNum != Integer.parseInt(split_data[5]) ? time : oppoTime));
	}
	
	@Override
	protected void sendMessage(String data, int capturePieceColor, boolean isMoveAck) throws Warning {
		board.send(data);
	}
	
	/**
	 * 
	 * 傳送初始盤面訊息給 Web。此為初始盤面時開啟盤面所呼叫的 function
	 * 
	 * @param data 初始盤面訊息，包含走步、時間、最終盤面和初始盤面。
	 * @param step_time 每個走步所花費的時間。
	 * @param total_spent_time 從遊戲開始到每個走步，各玩家所花費的時間累積。
	 * @param times 各玩家所剩餘的時間。
	 * @throws Warning 發生錯誤。
	 */
	
//	void sendInitBoard(String data, Integer[] step_time, Integer[] total_spent_time, int[] times) throws Warning{
//		String temp[] = data.split(",");
//		StringBuilder sb = new StringBuilder();
//		
//        // board + remain chess
//        for(int i = 0; i < 40; ++i){
//            sb.append(temp[i]);
//            sb.append(',');
//        }
//        // gameStarted
//        sb.append(temp[40]);
//        sb.append(',');
//		
//		// steps num
//		int step_num = Integer.parseInt(temp[41]);
//		sb.append(step_num);
//		// steps
//		for(int i = 0; i < step_num; i++){
//			sb.append(',');
//			sb.append(temp[42 + i]);
//			sb.append(',');
//			sb.append(step_time[i]);
//			sb.append(',');
//			sb.append(total_spent_time[i]);
//		}
//		// noEatFlip (noEatFlip Number always start from zero)
//		sb.append(',');
//		sb.append(0);
//		// first ply player time
//		sb.append(',');
//		sb.append(times[0]);
//		// second ply player time
//		sb.append(',');
//		sb.append(times[1]);
//		
//		board.send(sb.toString());
//	}
	
	/**
	 * 傳送遊戲結果給 Web 顯示。
	 * @param result 遊戲結果。
	 * @throws Warning 發生錯誤。
	 */
	void sendGameResult(String result) throws Warning {
		board.sendResult(result);
	}
	
	/**
	 * 傳送玩家的名稱給 Web 顯示。
	 * @param first_move_name 先手玩家名稱。
	 * @param second_move_name 後手玩家名稱。
	 */
	void setPlayerName(String first_move_name, String second_move_name){
		board.sendPlayerName(first_move_name, second_move_name);
	}
	
	/**
	 * 傳送詢問提和指令給盤面，讓使用者作出決定。
	 * @param semaphore 用來決定是否需要繼續詢問。
	 * @return 將使用者的回應包裝成 {@link client.Packet} 的 class，並傳送給伺服器。
	 * @throws Warning 發生錯誤。
	 */
	Packet askDrawOffer(BooleanHolder semaphore) throws Warning {
		board.send("drawOffer");
		
		// wait for data
		String result = "";
		do{
			try {
				result = board.recv(TIMEOUT);
			} catch(Warning w){
				switch(w.getType()){
				case WEBSERVER_RECV_TIMEOUT:
					if(semaphore.getValue() == false){
						return null;
					}else{
						continue;
					}
				default:
					Message.printError(ExceptionUtils.getStackTrace(w));
					return null;
				}
			}
		}while(result.equals(""));
		
		Message.printInfo("Action: %d", GST_Action.DrawOffer);
		return new Packet(GST_Action.DrawOffer, result.split(" ")[1]);
	}
	
	/**
	 * 傳送詢問是否悔棋的指令給 Web，讓使用者作出決定。
	 * @return 將使用者的回應包裝成 {@link client.Packet} 的 class，並傳送給伺服器。
	 * @throws Warning 發生錯誤。
	 */
	Packet askUndo() throws Warning{
		Packet p = new Packet();
		
		String result = board.recv();
		if(result.contains("undo")){
			if(isLocal){
				p.setAction(GST_Action.Undo);
				// substring the "undo " word
				p.setData(String.join(",", result.substring(5).split(" ")));
				return p;
			}else{
				// reject
				sendCommandResult("undo", 0);
			}
		}
		
		return p;
	}
	
	/**
	 * 因為在提出提和的過程中，提出方可以不等候對方回應，直接走步，因此也要接收提出方的走步資料。
	 * @param semaphore 用來決定是否需要繼續詢問。
	 * @return 將使用者的回應包裝成 {@link client.Packet} 的 class，並傳送給伺服器。
	 * @throws Warning 發生錯誤。
	 */
	Packet listenDrawOfferStep(BooleanHolder semaphore) throws Warning {
		// wait for data
		String result = "";
		do{
			try {
				result = board.recv(TIMEOUT);
			} catch(Warning w){
				switch(w.getType()){
				case WEBSERVER_RECV_TIMEOUT:
					if(semaphore.getValue() == false){
						return null;
					}else{
						continue;
					}
				default:
					Message.printError(ExceptionUtils.getStackTrace(w));
					return null;
				}
			}
		}while(result.equals(""));
		
		Packet p = new Packet();
		
		if(result.contains("MOV")){
			p.setAction(GST_Action.Move);
		}
		p.setData(result.split(":")[1]);
		return p;
	}
	
	/**
	 * 傳送某項指令給對方後，回傳給 Web 對方的回應。
	 * @param type 某項指令的類別。
	 * @param result 對方的回應。
	 * @throws Warning 發生錯誤。
	 */
	void sendCommandResult(String type, int result) throws Warning {
		board.send(String.format("%s %d", type, result));
	}
	
	/**
	 * 設置初始遊戲資料給 Search。
	 * @param setting 此房間的遊戲參數。
	 * @param midboard 遊戲初始盤面。
	 * @param myTime 我方初始時間。
	 * @param player_times 雙方初始剩餘時間。
	 * @param step_time 初始盤面各走步花費時間。
	 * @param total_spent_time 初始盤面從第一步到各走步雙方所累計的花費時間。
	 * @throws Warning 發生錯誤。
	 */
	@Override
	protected void setGameInfo(SettingFile setting, String midboard, int myTime, int[] player_times, Integer[] step_time, Integer[] total_spent_time) throws Warning {
		board.sendInitBoard(midboard, step_time, total_spent_time, player_times, 0);
	}
	
	/**
	 * {@inheritDoc}
	 * 設定玩家編號。
	 * @param playerNum 玩家編號。
	 */
	@Override
	protected void gameStart(int playerNum, boolean firstMove) throws Warning {
		this.playerNum = playerNum;
	}
	
	/**
	 * {@inheritDoc}
	 * 停止 Search 運行。
	 */
	@Override
	protected void stopSearch() {
		finish = true;
		board.stopWebServer();
	}
	
	/**
	 * 傳送重新開始一局的指令給 Web，用來重整盤面。
	 * @throws Warning 發生錯誤。
	 */
	void restartBoard() throws Warning{
		board.restartBoard();
	}
	
	/**
	 * 設置此玩家的顏色給 Search。
	 * @param color 玩家顏色。
	 */
	void setColor(int color){//改玩家
		switch(color){
		case 0:
			this.color = "0";
			break;
		case 1:
			this.color = "1";
			break;
		case 2:
			this.color = "unknown";
			break;
		}
	}
	
	/**
	 * 因為某些指令需要判別對方是否是人類還是 AI，因此需要設置對方的玩家類別。
	 * @param isHuman 是否是人類。
	 */
	void setOppoisHuman(boolean isHuman){
		this.oppoIsHuamn = isHuman;
	}
	
	/**
	 * 設置遊戲是否結束。
	 * @param isFinish 是否結束。
	 */
	void setisFinish(BooleanHolder isFinish){
		this.isFinish = isFinish;
	}
	
	/**
	 * 設定新的 instance。
	 * @param 新的 instance。
	 */
	void resetBoard(GST_Board new_board){
		this.board = new_board;
	}
	
	/**
	 * 
	 * 取得和 Search 的所有互動字串。
	 * 
	 * @return 和 Search 的所有互動字串。
	 */
	protected String getInteractiveLog(){
		return ""; // Human mode is useless
	}
}