package gst;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.Socket;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import org.apache.commons.lang3.exception.ExceptionUtils;

import client.Message;
import client.Packet;
import client.SettingFile;
import client.Version;
import client.Warning;
import client.Warning.Type;

/**
 * 處理有關 MGTP 連接模式下的連接事宜。
 * 
 * @author ldslds449
 * @see gst.SearchConnect
 */
class MGTP extends SearchConnect{
	private Process search;
	private String filePath;
	
	private Thread readStdin = null;
	private Thread readStderr = null;
	
	private BufferedReader stdinIS;
	private BufferedReader stderrIS;
	
	private StringBuilder commandLogSB;
	
	private int color;
	private String reply = "";
	private boolean replyResult;
	private boolean isAlive = true;
	
	private int playerNum;
	private GST_Room room;
	
	/**
	 * MGTP 的指令列表及代碼。
	 * @author ldslds449
	 *
	 */
//	private static enum Command{
//		protocol_version, // 0
//		name,
//		version,
//		known_command,
//		list_commands,
//		quit, // 5
//		boardsize,
//		reset_board,
//		num_repetition,
//		num_moves_to_draw,
//		move, // 10
//		flip,
//		genmove,
//		game_over,
//		ready,
//		time_setting, // 15
//		time_left,
//		showboard,
//		init_board};
	
	/**
	 * constructor，依照傳入的檔案路徑啟動 Search，設定讀取標準輸出和標準錯誤輸出的 Thread。
	 * @param filePath Search 檔案路徑。
	 * @param arg 程式啟動參數。
	 * @param room 房間頁面的 instance，用來輸出標準輸出和標準錯誤輸出的 log。 
	 * @throws Warning 發生錯誤。
	 */
	MGTP(String filePath, String arg, GST_Room room) throws Warning{
		
		this.filePath = filePath;
		this.color = 2;
		this.commandLogSB = new StringBuilder();
		this.room = room;
		
		try {
			List<String> commands = new ArrayList<String>();
			commands.add(this.filePath);
			if(!arg.matches(" *")){ // args
				String arg_modify = arg.trim().replaceAll(" +", " ");
				commands.addAll(Arrays.asList(arg_modify.split(" ")));
			}
			
			ProcessBuilder p = new ProcessBuilder(commands);
			p.directory(Paths.get(filePath).toFile().getAbsoluteFile().getParentFile());
			Message.printInfo(">> Call Command: %s", String.join(" ", p.command()));
			search = p.start();
		} catch (IOException e) {
			Message.printError("ERROR: START SEARCH IN MGTP");
			Message.printError("FilePath: %s\n", filePath);
			Message.printError(ExceptionUtils.getStackTrace(e));
			
			stopSearch();
			
			throw new Warning(Type.SEARCH_CONNECT_ERROR, String.format("CONNECT TO SEARCH ERROR%s>>%s", System.getProperty("line.separator"), filePath));
		}



		readStdin = new Thread(() -> {
			stdinIS = new BufferedReader(new InputStreamReader(search.getInputStream()));
			try {
				String temp = null;
				while(!finish && search.isAlive()){
					// check if there is data to read
					while(!finish && !stdinIS.ready()){
						// check search if is alive
						if(!search.isAlive()){
							synchronized(readStdin){
								isAlive = false;
								readStdin.notify();
							}
							break;
						}
						// wait
						if(STDOUT_READ_DELAY != 0){
							try {
								Thread.sleep(STDOUT_READ_DELAY);
							} catch (InterruptedException e) {
								Message.printError(ExceptionUtils.getStackTrace(e));
							}
						}else{
							STDOUT_READ_DELAY = STDOUT_READ_DELAY_INC;
						}
					}
					if(!isAlive || finish) break;

					// read data
					temp = stdinIS.readLine();
					if(temp == null || finish) break;
					STDOUT_READ_DELAY = 0;
					synchronized(MGTP.this){
						if(temp.indexOf("MOV") != -1 || temp.indexOf("SET") != -1){
							synchronized(readStdin){
								reply = temp;
								replyResult = true;
								readStdin.notify();
							}
						}
//						if(temp.startsWith("=") && temp.substring(1).split(" ")[0].matches("^[0-9]*$")){
//							synchronized(readStdin){
//								reply = temp.substring(1);
//								replyResult = true;
//								readStdin.notify();
//							}
//						}else if(temp.startsWith("?") && temp.substring(1).split(" ")[0].matches("^[0-9]*$")){
//							synchronized(readStdin){
//								reply = temp.substring(1);
//								replyResult = false;
//								readStdin.notify();
//							}
//						}

						if(room != null){
							room.setConnectInfo(String.valueOf(temp));
						}else{
							Message.printCLI(String.valueOf(temp));
						}
						connect_Info.append(String.format("%s%s", String.valueOf(temp), System.lineSeparator()));
					}
				}
				synchronized(readStdin){
					isAlive = false;
					readStdin.notify();
				}
				Message.printInfo("Close Search STDOUT Stream");
			} catch (IOException e) {
				Message.printError("#ERROR MGTP READ STDIN STREAM");
				Message.printError(ExceptionUtils.getStackTrace(e));
			}
	    });
		readStdin.start();

		readStderr = new Thread(() -> {
			stderrIS = new BufferedReader(new InputStreamReader(search.getErrorStream()));
			try {
				String temp = null;
				while(!finish && (temp = stderrIS.readLine()) != null){
					synchronized(MGTP.this){
						if(room != null){
							room.setErrorInfo(String.valueOf(temp));
						}else{
							Message.printErrCLI(String.valueOf(temp));
						}
						error_Info.append(String.format("%s%s", String.valueOf(temp), System.lineSeparator()));
					}
					// wait
					if(STDERR_READ_DELAY != 0){
						try {
							Thread.sleep(STDERR_READ_DELAY);
						} catch (InterruptedException e) {
							Message.printError(ExceptionUtils.getStackTrace(e));
						}
					}
				}
				Message.printInfo("Close Search STDERR Stream");
			} catch (IOException e) {
				Message.printError("#ERROR MGTP READ ERROR STREAM");
				Message.printError(ExceptionUtils.getStackTrace(e));
			}
	    });
		readStderr.start();
	}
	
	/**
	 * {@inheritDoc}
	 * 要求 Search 產生走步。
	 * @return Packet class({@link client.Packet}) 的封包，內含動作代碼和走步。
	 */
	@Override
	protected Packet askMove(String ghostCodeInfo) throws Warning {
		
		// before the move of the opponent, add separation line
		stdout_lock.writeLock().lock();
		stderr_lock.writeLock().lock();
		stdout_stamp.add(getConnectInfo().split("\r\n|\r|\n").length);
		stderr_stamp.add(getConnectInfo().split("\r\n|\r|\n").length);
		stdout_lock.writeLock().unlock();
		stderr_lock.writeLock().unlock();
		
		send(String.format("%s", "MOV?" + ghostCodeInfo + "\r\n"));
		String data = recv();
		String[] data_split = data.split(":");
		
		Packet p = new Packet();
		// judge if the flip or move
		if(data_split[0].indexOf("MOV") != -1){
//			Message.printInfo("Enter askmove");
			p.setAction(GST_Action.Move);
			p.setData(String.format("%s", data_split[1]));
		}
		
		return p;
		
	}
	
	/**
	 * {@inheritDoc}
	 * 傳送走步給 Search。
	 * @param data Server 傳送過來的走步完整資料(包含走步、時間等)。
	 * @param noEatFlip 無吃翻數字。
	 * @param step_time 當前走步所花費的時間。
	 * @param total_spent_time 此走步玩家到當前走步所花費的時間累計。
	 * @param time 我方玩家所剩餘的時間。
	 * @param oppoTime 對手剩餘時間。
	 * @param isLast 這步是否為最後一步，避免 Search 多走一步。
	 */
	@Override
	protected void sendMove(String data, int capturePieceColor, int noEatFlip, int step_time, int total_spent_time, int time, int oppoTime, boolean isLast) throws Warning {
		if(capturePieceColor == 0) {
			send("OK \r\n");
		}else if(capturePieceColor == 1) {
			send("OKR\r\n");
		}else if(capturePieceColor == 2){
			send("OKB\r\n");
		}
	}
	
	@Override
	protected void sendMessage(String data, int capturePieceColor, boolean isMoveAck) throws Warning {
		if(isMoveAck) {
			if(capturePieceColor == 0) {
				send(data + " \r\n");
			}else if(capturePieceColor == 1) {
				send(data + "R\r\n");
			}else if(capturePieceColor == 2){
				send(data + "B\r\n");
			}
		}else {
			send(data + "\r\n");
		}
	}
	@Override
	protected Packet askSetRed() throws Warning {
		// before the move of the opponent, add separation line
		stdout_lock.writeLock().lock();
		stderr_lock.writeLock().lock();
		stdout_stamp.add(getConnectInfo().split("\r\n|\r|\n").length);
		stderr_stamp.add(getConnectInfo().split("\r\n|\r|\n").length);
		stdout_lock.writeLock().unlock();
		stderr_lock.writeLock().unlock();
		
		send("SET?\r\n");
		String data = recv();
		String[] data_split = data.split(":");
		
		Packet p = new Packet();
		if(data_split[0].indexOf("SET") != -1){
			p.setAction(GST_Action.SetRed);
			p.setData(String.format("%s", data_split[1]));
		}
		
		return p;
	}
	
	/**
	 * 設置初始遊戲資料給 Search。
	 * @param setting 此房間的遊戲參數。
	 * @param midboard 遊戲初始盤面。
	 * @param myTime 我方初始時間。
	 * @param player_times 雙方初始剩餘時間。
	 * @param step_time 初始盤面各走步花費時間。
	 * @param total_spent_time 初始盤面從第一步到各走步雙方所累計的花費時間。
	 * @throws Warning 發生錯誤。
	 */
	@Override
	protected void setGameInfo(SettingFile setting, String midboard, int myTime, int[] player_times, Integer[] step_time, Integer[] total_spent_time) throws Warning {
//		String[] temp = midboard.split(",");
//		
//		StringBuilder sb = new StringBuilder();
//		sb.append("/start ");
//
//		int step_size = Integer.parseInt(temp[41]);
//		int time_size = Integer.parseInt(temp[42 + step_size]);
//		
//		// initial board
//		for(int i = 0; i < 8; ++i){
//			for(int j = 0; j < 4; ++j){
//				sb.append(pieceKind[Integer.parseInt(temp[49 + step_size + time_size + j + (7-i)*4])]);
//				sb.append(',');
//			}
//		}
//		// initial remain piece
//		for(int i = 0; i < 14; i++){
//			sb.append(temp[49 + step_size + time_size + 32 + i]); // initial number
//			sb.append(',');
//		}
//		// initial history
//		sb.append(step_size); // zero of history move
//		sb.append(',');
//		for(int i = 48; i < 48 + step_size; i++){
//			sb.append(temp[i]);
//			sb.append(',');
//		}
//		// initial time
//		sb.append(myTime);
//		
//		send(sb.toString());
//		// wait for search ready
//		String recv_data = recv();
//		if(!recv_data.equals("/start")){
//			Message.printError("#ERROR SEARCH START NOT YET");
//			Message.printError("#%s", sb.toString());
//			
//			stopSearch();
//			throw new Warning(Type.SEARCH_FORMAT_ERROR, "RECV DATA NEED REQUIRE \"/start\"");
//		}
	}
	
	/**
	 * {@inheritDoc}
	 * 設定玩家編號。
	 * @param playerNum 玩家編號。
	 */
	@Override
	protected void gameStart(int playerNum, boolean firstMove) throws Warning {
		this.playerNum = playerNum;
		
//		if(!firstMove){
//			stdout_lock.writeLock().lock();
//			stderr_lock.writeLock().lock();
//			stdout_stamp.add(getConnectInfo().split("\r\n|\r|\n").length);
//			stderr_stamp.add(getConnectInfo().split("\r\n|\r|\n").length);
//			stdout_lock.writeLock().unlock();
//			stderr_lock.writeLock().unlock();
//		}
	}
	
	/**
	 * {@inheritDoc}
	 * 停止 Search 運行。
	 */
	@Override
	protected void stopSearch() {
		finish = true;
		
		if(search != null && search.isAlive()){
			try{
				send("/exit");
			}catch(Warning w){}
		}
		
		// close search
		if(search != null && search.isAlive()){
			try {
				search.destroyForcibly().waitFor();
			} catch (InterruptedException e) {}
		}
		
		if(readStdin != null && readStdin.isAlive()){
			readStdin.interrupt();
			try {
				stdinIS.close();
			} catch (IOException e) {
				Message.printError(ExceptionUtils.getStackTrace(e));
			}
		}
		
		if(readStderr != null && readStderr.isAlive()){
			readStderr.interrupt();
			try {
				stderrIS.close();
			} catch (IOException e) {
				Message.printError(ExceptionUtils.getStackTrace(e));
			}
		}

	}
	
	/**
	 * 傳送資料給 Search。
	 * @param data 要傳送的資料。
	 * @throws Warning 發生錯誤。
	 */
	private void send(String data) throws Warning{
		Message.printInfo("MGTP Send: %s", data);
		commandLogSB.append(String.format("%s%n", data));
		if(room != null){
			room.setConnectInfo(String.format("[Client Send: %s]", data));
			room.setInteractiveLog(String.format("%s%n", data));
		}else{
			Message.printCLI(String.format("[Client Send: %s]", data));
		}
		
		// clean the reply
		// because one command need to receive one response
		reply = "";
		
		try {
			BufferedWriter BW = new BufferedWriter(new OutputStreamWriter(search.getOutputStream()));
			BW.write(data);
			BW.write(System.getProperty("line.separator"));
			
			BW.flush();
		} catch (IOException e) {
			Message.printError("#ERROR STDOUT SEND DATA TO SEARCH");
			Message.printError("FilePath: %s\n", filePath);
			Message.printError(ExceptionUtils.getStackTrace(e));
			
			stopSearch();
			
			throw new Warning(Type.SEARCH_SEND_ERROR, "STDOUT SEND DATA TO SEARCH ERROR");
		}
		
	}
	
	/**
	 * 從 Search 接收資料。
	 * @param ID 指令代號，MGTP 模式下，因為接收到的指令代號要與傳送過去的對應，因此要確認接收的指令代號。
	 * @return 從 Search 接收到的資料
	 * @throws Warning 發生錯誤。
	 */
	private String recv() throws Warning {
		Message.printInfo("Entering recv method");
		try {
	        synchronized (readStdin) {
	            while (reply.length() <= 0 && isAlive) {
	                readStdin.wait();
	            }
	        }
	    } catch (InterruptedException e) {
	        Message.printError("#ERROR STDIN RECV DATA FROM SEARCH");
	        Message.printError("FilePath: %s\n", filePath);
	        Message.printError(ExceptionUtils.getStackTrace(e));

	        stopSearch();
	        throw new Warning(Type.SEARCH_RECV_ERROR, "STDIN RECV DATA FROM SEARCH ERROR");
	    }

	    Message.printInfo("MGTP Recv: %s", reply);
	    commandLogSB.append(String.format(">>%s%n", reply));
	    if (room != null) {
	        room.setConnectInfo(String.format("[Client Recv: %s]", reply));
	        room.setInteractiveLog(String.format(">>%s%n", reply));
	    } else {
	        Message.printCLI(String.format("[Client Recv: %s]", reply));
	    }

	    if (!isAlive) {
	        Message.printError("#ERROR SEARCH CLOSE");
	        Message.printError("FilePath: %s\n", filePath);

	        stopSearch();
	        throw new Warning(Type.SEARCH_NO_RESPONSE, "SEARCH CLOSE ERROR");
	    }
	    if (reply.endsWith("\r\n")) {
            reply = reply.substring(0, reply.length() - 2);
        }
	    return reply;
	}

	

	/**
	 * 
	 * 設定我方顏色。
	 * 
	 * @param color 我方顏色。
	 */
	void setColor(int color){
		this.color = color;
	}
	
	/**
	 * 
	 * 將顏色從數字改成字串。
	 * 
	 * @param color 顏色。
	 * @return color 顏色字串。
	 */
//	String getColor(int color){
//		if(color == 0){
//			return "red";
//		}else if(color == 1){
//			return "black";
//		}else if(color == 2){
//			return "unknown";
//		}
//		return "";
//	}
//	
	/**
	 * 
	 * 取得程式名稱。
	 * 
	 * @return 程式名稱。
	 * @throws Warning 發生錯誤。
	 */
//	String getSearchName() throws Warning{
//		// ask Search Name
//		send(String.format("%d %s", Command.name.ordinal(), Command.name.name()));
//		return recv(Command.name.ordinal()).split(" ")[1];
//	}
	
	/**
	 * 
	 * 取得程式版本。
	 * 
	 * @return 程式版本。
	 * @throws Warning 發生錯誤。
	 */
//	String getSearchVersion() throws Warning{
//		// ask Search Version
//		send(String.format("%d %s", Command.version.ordinal(), Command.version.name()));
//		return recv(Command.version.ordinal()).split(" ")[1];
//	}
	
	/**
	 * 
	 * 取得和 Search 的所有互動字串。
	 * 
	 * @return 和 Search 的所有互動字串。
	 */
	protected String getInteractiveLog(){
		return commandLogSB.toString();
	}
}
