package gst;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.ServerSocket;
import java.net.Socket;
import java.nio.charset.Charset;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import org.apache.commons.lang3.exception.ExceptionUtils;

import client.Message;
import client.Packet;
import client.SettingFile;
import client.Warning;
import client.Warning.Type;

/**
 * 處理有關 Ponder 連接模式下的連接事宜。
 * 
 * @author ldslds449
 * @see gst.SearchConnect
 */
class Ponder extends SearchConnect{
	
	private Socket socket;
	private Process search;
	
	private Thread readStdin = null;
	private Thread readStderr = null;
	
	private BufferedReader stdinIS;
	private BufferedReader stderrIS;
	
	private StringBuilder commandLogSB;
	
	private int playerNum;
	private int color;
	
	private GST_Room room;
	
	/**
	 * constructor，依照傳入的檔案路徑啟動 Search，設定讀取標準輸出和標準錯誤輸出的 Thread。
	 * @param filePath Search 檔案路徑。
	 * @param arg 程式啟動參數。
	 * @param room 房間頁面的 instance，用來輸出標準輸出和標準錯誤輸出的 log。 
	 * @throws Warning 發生錯誤。
	 */
	Ponder(String filePath, String arg, GST_Room room) throws Warning{
		
		this.color = 2;
		this.commandLogSB = new StringBuilder();
		this.room = room;
		
		ServerSocket serverSocket = null;
		try{
			serverSocket = new ServerSocket(0);
			Message.printInfo("Ponder IP: %s Port: %d", "127.0.0.1", serverSocket.getLocalPort());
			
			if(room != null){room.changeTitle(String.format("Ponder IP: 127.0.0.1 Port: %d", serverSocket.getLocalPort()));}
			else{Message.printCLI(String.format("Ponder IP: 127.0.0.1 Port: %d", serverSocket.getLocalPort()));}
		}catch(IOException e){
			Message.printError("#ERROR CREATE SOCKET");
			Message.printError("Ponder IP: %s", "127.0.0.1");
			Message.printError(ExceptionUtils.getStackTrace(e));
			
			stopSearch();
			
			throw new Warning(Type.SEARCH_CONNECT_ERROR, "SOCKET CREATE ERROR");
		}
		
		try {
			List<String> commands = new ArrayList<String>();
			commands.add(filePath);
			Message.printInfo(">> arg: %s", arg);
			if(!arg.matches(" *")){ // args
				String arg_modify = arg.trim().replaceAll(" +", " ");
				commands.addAll(Arrays.asList(arg_modify.split(" ")));
			}
			commands.add("127.0.0.1"); // ip
			commands.add(Integer.toString(serverSocket.getLocalPort())); // port
			
			
			ProcessBuilder p = new ProcessBuilder(commands);
//			if(arg.matches(" *")){
//				p = new ProcessBuilder(Paths.get(filePath).toFile().getCanonicalPath(), 
//						"127.0.0.1", Integer.toString(serverSocket.getLocalPort()));
//			}else{
//				p = new ProcessBuilder(Paths.get(filePath).toFile().getCanonicalPath(), 
//						"127.0.0.1", Integer.toString(serverSocket.getLocalPort()), arg);
//			}
			p.directory(Paths.get(filePath).toFile().getAbsoluteFile().getParentFile());
			Message.printInfo(">> Call Command: %s", String.join(" ", p.command()));
			search = p.start();
			// open search in cmd.exe
			//search = new ProcessBuilder("cmd.exe", "/K", "Start", filePath, "127.0.0.1", Integer.toString(serverSocket.getLocalPort())).start();
		} catch (IOException e) {
			Message.printError("#ERROR CALL SEARCH EXE");
			Message.printError("#FilePath: %s", filePath);
			Message.printError("Ponder IP: %s, Port: %d", "127.0.0.1", serverSocket.getLocalPort());
			Message.printError(ExceptionUtils.getStackTrace(e));
			
			stopSearch();
			
			try{serverSocket.close();}catch(IOException ignore){}
			throw new Warning(Type.SEARCH_NO_FOUND, String.format("CAN NOT FOUND SEARCH%s>>%s", System.getProperty("line.separator"), filePath));
		}
		
		// read standard output
		/* we need to read standard output and standard error (below is java doc)
		 * Because some native platforms only provide limited buffer size for standard input and output streams, 
		 * failure to promptly write the input stream or read the output stream of the subprocess may cause the subprocess to block, and even deadlock.
		 */
		readStdin = new Thread(() -> {
			stdinIS = new BufferedReader(new InputStreamReader(search.getInputStream()));
			try {
				String temp = null;
				while((temp = stdinIS.readLine()) != null){
					synchronized(Ponder.this){
						if(room != null){
							room.setConnectInfo(String.valueOf(temp));
						}else{
							Message.printErrCLI(String.valueOf(temp));
						}
						connect_Info.append(String.format("%s%s", String.valueOf(temp), System.lineSeparator()));
					}
					// wait
					try {
						Thread.sleep(STDOUT_READ_DELAY);
					} catch (InterruptedException e) {
						Message.printError(ExceptionUtils.getStackTrace(e));
					}
				}
				Message.printInfo("Close Search STDOUT Stream");
			} catch (IOException e) {
				Message.printError("#ERROR PONDER READ STDIN STREAM");
				Message.printError(ExceptionUtils.getStackTrace(e));
			}
	    });
		readStdin.setPriority(stdThreadPriority);
		readStdin.start();
		
		readStderr = new Thread(() -> {
			stderrIS = new BufferedReader(new InputStreamReader(search.getErrorStream()));
			try {
				String temp = null;
				while((temp = stderrIS.readLine()) != null){
					synchronized(Ponder.this){
						if(room != null){
							room.setErrorInfo(String.valueOf(temp));
						}else{
							Message.printErrCLI(String.valueOf(temp));
						}
						error_Info.append(String.format("%s%s", String.valueOf(temp), System.lineSeparator()));
					}
					// wait
					try {
						Thread.sleep(STDERR_READ_DELAY);
					} catch (InterruptedException e) {
						Message.printError(ExceptionUtils.getStackTrace(e));
					}
				}
				Message.printInfo("Close Search STDERR Stream");
			} catch (IOException e) {
				Message.printError("#ERROR PONDER READ ERROR STREAM");
				Message.printError(ExceptionUtils.getStackTrace(e));
			}
	    });
		readStderr.setPriority(errThreadPriority);
		readStderr.start();
		
		// accept socket
		try {
			socket = serverSocket.accept();
			serverSocket.close();
		} catch (IOException e) {
			Message.printError("#ERROR ACCEPT SOCKET CONNECT");
			Message.printError("#FilePath: %s", filePath);
			Message.printError("Ponder IP: %s, Port: %d", "127.0.0.1", serverSocket.getLocalPort());
			Message.printError(ExceptionUtils.getStackTrace(e));
			
			stopSearch();
			
			throw new Warning(Type.SEARCH_CONNECT_ERROR, String.format("CLIENT ACCEPT SEARCH ERROR%s>>IP: 127.0.0.1 Port: %d", System.getProperty("line.separator"), serverSocket.getLocalPort()));
		}
	}
	
	/**
	 * {@inheritDoc}
	 * 要求 Search 產生走步。
	 * @return Packet class({@link client.Packet}) 的封包，內含動作代碼和走步。
	 */
	@Override
	protected Packet askMove(String ghostCodeInfo) throws Warning {
		
		// before the move of the opponent, add separation line
		stdout_lock.writeLock().lock();
		stderr_lock.writeLock().lock();
		stdout_stamp.add(getConnectInfo().split("\r\n|\r|\n").length);
		stderr_stamp.add(getConnectInfo().split("\r\n|\r|\n").length);
		stdout_lock.writeLock().unlock();
		stderr_lock.writeLock().unlock();
		
		send(String.format("%s", "MOV?" + ghostCodeInfo + "\r\n"));
		String data = recv();
		String[] data_split = data.split(":");
		
		Packet p = new Packet();
		// judge if the flip or move
		if(data_split[0].indexOf("MOV") != -1){
			p.setAction(GST_Action.Move);
			p.setData(String.format("%s", data_split[1]));
		}
		
		return p;
	}
	
	@Override
	protected Packet askSetRed() throws Warning {
		// before the move of the opponent, add separation line
		stdout_lock.writeLock().lock();
		stderr_lock.writeLock().lock();
		stdout_stamp.add(getConnectInfo().split("\r\n|\r|\n").length);
		stderr_stamp.add(getConnectInfo().split("\r\n|\r|\n").length);
		stdout_lock.writeLock().unlock();
		stderr_lock.writeLock().unlock();
		
		send("SET?\r\n");
		String data = recv();
		String[] data_split = data.split(":");
		
		Packet p = new Packet();
		if(data_split[0].indexOf("SET") != -1){
			p.setAction(GST_Action.SetRed);
			p.setData(String.format("%s", data_split[1]));
		}
		
		return p;
	}

	/**
	 * {@inheritDoc}
	 * 傳送走步給 Search。
	 * @param data Server 傳送過來的走步完整資料(包含走步、時間等)。
	 * @param noEatFlip 無吃翻數字。
	 * @param step_time 當前走步所花費的時間。
	 * @param total_spent_time 此走步玩家到當前走步所花費的時間累計。
	 * @param time 我方玩家所剩餘的時間。
	 * @param oppoTime 對手剩餘時間。
	 * @param isLast 這步是否為最後一步，避免 Search 多走一步。
	 */
	@Override
	protected void sendMove(String data, int capturePieceColor, int noEatFlip, int step_time, int total_spent_time, int time, int oppoTime, boolean isLast) throws Warning {
//		String[] temp = data.split(",");
//		
//		int player = (Integer.parseInt(temp[3]) == 1 ? 2 : 1);
//		// before the move of the opponent, add separation line
//		if(player == playerNum){
//			stdout_lock.writeLock().lock();
//			stderr_lock.writeLock().lock();
//			stdout_stamp.add(getConnectInfo().split("\r\n|\r|\n").length);
//			stderr_stamp.add(getConnectInfo().split("\r\n|\r|\n").length);
//			stdout_lock.writeLock().unlock();
//			stderr_lock.writeLock().unlock();
//		}
//		
//		StringBuilder sb = new StringBuilder();
//		String move = temp[0];
//		if(move.charAt(2) == '('){
//			sb.append("/flip ");
//		}else{
//			sb.append("/move ");
//		}
//		// add location
//		sb.append(move);
//		// add time
//		sb.append(' ');
//		sb.append(time);
//		// add isLast
//		sb.append(' ');
//		sb.append((isLast ? 1 : 0));
//		
		// send Data
//		send(sb.toString());
	}
	
	@Override
	protected void sendMessage(String data, int capturePieceColor, boolean isMoveAck) throws Warning {
		if(isMoveAck) {
			if(capturePieceColor == 0) {
				send(data + " \r\n");
			}else if(capturePieceColor == 1) {
				send(data + "R\r\n");
			}else if(capturePieceColor == 2){
				send(data + "B\r\n");
			}
		}else {
			send(data + "\r\n");
		}
	}
	
	/**
	 * 設置初始遊戲資料給 Search。
	 * @param setting 此房間的遊戲參數。
	 * @param midboard 遊戲初始盤面。
	 * @param myTime 我方初始時間。
	 * @param player_times 雙方初始剩餘時間。
	 * @param step_time 初始盤面各走步花費時間。
	 * @param total_spent_time 初始盤面從第一步到各走步雙方所累計的花費時間。
	 * @throws Warning 發生錯誤。
	 */
	@Override
	protected void setGameInfo(SettingFile setting, String midboard, int myTime, int[] player_times, Integer[] step_time, Integer[] total_spent_time) throws Warning{
		String[] temp = midboard.split(",");
		
		StringBuilder sb = new StringBuilder();
		sb.append("/start ");

		int step_size = Integer.parseInt(temp[41]);
		int time_size = Integer.parseInt(temp[42 + step_size]);
		
		// initial board
		for(int i = 0; i < 8; ++i){
			for(int j = 0; j < 4; ++j){
				sb.append(pieceKind[Integer.parseInt(temp[49 + step_size + time_size + j + (7-i)*4])]);
				sb.append(',');
			}
		}
		// initial remain piece
		for(int i = 0; i < 14; i++){
			sb.append(temp[49 + step_size + time_size + 32 + i]); // initial number
			sb.append(',');
		}
		// initial history
		sb.append(step_size); // zero of history move
		sb.append(',');
		for(int i = 48; i < 48 + step_size; i++){
			sb.append(temp[i]);
			sb.append(',');
		}
		// initial time
		sb.append(myTime);
		
		send(sb.toString());
		// wait for search ready
		String recv_data = recv();
		if(!recv_data.equals("/start")){
			Message.printError("#ERROR SEARCH START NOT YET");
			Message.printError("#%s", sb.toString());
			
			stopSearch();
			throw new Warning(Type.SEARCH_FORMAT_ERROR, "RECV DATA NEED REQUIRE \"/start\"");
		}
	}
	
	/**
	 * {@inheritDoc}
	 * 設定玩家編號。
	 * @param playerNum 玩家編號。
	 */
	@Override
	protected void gameStart(int playerNum, boolean firstMove) throws Warning {
		this.playerNum = playerNum;
	}
//	@Override
//	protected void gameStart(int playerNum, boolean firstMove) throws Warning{
//		this.playerNum = playerNum;
//		StringBuilder sb = new StringBuilder();
//		
//		// send turn and color
//		sb.append("/turn");
//		// isfirst
//		sb.append(String.format(" %d ", (firstMove ? 1 : 0)));
//		// color (2: unknown)
//		sb.append(color);
//		
//		
//		send(sb.toString());
//		// wait for search ready
//		String recv_data = recv();
//		if(!recv_data.equals("/turn")){
//			Message.printError("#ERROR SEARCH GET TURN NOT YET");
//			Message.printError("#%s", sb.toString());
//			
//			stopSearch();
//			throw new Warning(Type.SEARCH_FORMAT_ERROR, "RECV DATA NEED REQUIRE \"/turn\"");
//		}
//		
//		if(!firstMove){
//			stdout_lock.writeLock().lock();
//			stderr_lock.writeLock().lock();
//			stdout_stamp.add(getConnectInfo().split("\r\n|\r|\n").length);
//			stderr_stamp.add(getConnectInfo().split("\r\n|\r|\n").length);
//			stdout_lock.writeLock().unlock();
//			stderr_lock.writeLock().unlock();
//		}
//	}
	
	/**
	 * 從 Search 接收資料。
	 * @return 從 Search 接收到的資料
	 * @throws Warning 發生錯誤。
	 */
	private String recv() throws Warning{
		String result = "";
		try {
//			DataInputStream DIS = new DataInputStream(socket.getInputStream());
//			
//			int packet_length_1 = DIS.read();
//			int packet_length_2 = DIS.read();
//			
//			int packet_length = packet_length_1 + packet_length_2 * 128;
//			
//			Message.printInfo("DIS packet_length_1 = %d, packet_length_2 = %d, packet_length = %d", packet_length_1, packet_length_2, packet_length);
//			
//			byte[] data = new byte[packet_length];
//			int data_len = 0;
//			do{
//				int read_num = DIS.read(data, data_len, packet_length);
//				data_len += read_num;
//				packet_length -= read_num;
//			}while(packet_length > 0);
			InputStream inputStream = socket.getInputStream();
			StringBuilder resultBuilder = new StringBuilder();
			byte[] buffer = new byte[1024];
			while(!(result.indexOf("\r\n") > 0)){
				int len = inputStream.read(buffer);
				if(len == -1){
					Message.printError("#ERROR SOCKET RECV DATA FROM SEARCH");
					throw new RuntimeException("channel is not opend.");
				}
				resultBuilder.append(new String(buffer, 0, len));
				result = resultBuilder.toString();
			}
			
			if (result.endsWith("\r\n")) {
                result = result.substring(0, result.length() - 2);
            }
			
		} catch (IOException e) {
			Message.printError("#ERROR SOCKET RECV DATA FROM SEARCH");
			Message.printError("#IP: %s, Port: %d", "127.0.0.1", socket.getPort());
			Message.printError(ExceptionUtils.getStackTrace(e));
			
			// close search
			stopSearch();
			
			throw new Warning(Type.SEARCH_RECV_ERROR, "SEARCH RECV DATA ERROR");
		}
		
		Message.printInfo("Ponder Recv: %s", result);
		commandLogSB.append(String.format(">>%s%n", result));
		if(room != null){
			room.setConnectInfo(String.format("[Client Recv: %s]", result));
			room.setInteractiveLog(String.format(">>%s%n", result));
		}else{
			Message.printCLI(String.format("[Client Recv: %s]", result));
		}
		
		return result;
	}
	
	/**
	 * 傳送資料給 Search。
	 * @param data 要傳送的資料。
	 * @throws Warning 發生錯誤。
	 */
	private void send(String data) throws Warning{
		Message.printInfo("Ponder Send: %s", data);
		commandLogSB.append(String.format("%s%n", data));
		if(room != null){
			room.setConnectInfo(String.format("[Client Send: %s]", data));
			room.setInteractiveLog(String.format("%s%n", data));
		}else{
			Message.printCLI(String.format("[Client Send: %s]", data));
		}
		
		try {
			DataOutputStream DOS = new DataOutputStream(socket.getOutputStream());
			
			StringBuilder sb = new StringBuilder();
			// data length
//			char data_length_1 = (char)(data.getBytes().length % 128);
//			char data_length_2 = (char)(data.getBytes().length / 128);
//			sb.append(data_length_1);
//			sb.append(data_length_2);
			sb.append(data);
			
			DOS.write(sb.toString().getBytes());
			
			DOS.flush();
		} catch (IOException e) {
			Message.printError("#ERROR SOCKET SEND DATA TO SEARCH");
			Message.printError("#IP: %s, Port: %d", "127.0.0.1", socket.getPort());
			Message.printError(ExceptionUtils.getStackTrace(e));

			// close search
			stopSearch();
			
			throw new Warning(Type.SEARCH_SEND_ERROR, "SEARCH SEND DATA ERROR");
		}
	}
	/**
	 * 設置此玩家的顏色給 Search。
	 * @param color 玩家顏色。
	 */
	
	void setColor(int color){
		this.color = color;
	}
	
	/**
	 * {@inheritDoc}
	 * 停止 Search 運行。
	 */
	@Override
	protected void stopSearch() {
		finish = true;
		if(search != null && search.isAlive()){
			try {
				send("/exit");
			} catch (Warning e) {}
		}
		
		Timer wait_for_close = new Timer();
		wait_for_close.schedule(new TimerTask(){
			@Override
			public void run() {
				if(search != null && search.isAlive()){
					try{
						search.destroyForcibly().waitFor();
					} catch (InterruptedException e) {
						Message.printError(ExceptionUtils.getStackTrace(e));
					}
				}
			}
		}, 500);
		
		// close search
		if(search != null && search.isAlive()){
			try{
				search.waitFor();
			} catch (InterruptedException e) {
				Message.printError(ExceptionUtils.getStackTrace(e));
			}
		}
		
		if(socket != null){
			try {
				socket.close();
			} catch (IOException e) {
				Message.printError(ExceptionUtils.getStackTrace(e));
			}
		}
		
		if(readStdin != null && stdinIS != null){
			try {
				stdinIS.close();
			} catch (IOException e) {
				Message.printError(ExceptionUtils.getStackTrace(e));
			}
		}
		
		if(readStderr != null && stderrIS != null){
			try {
				stderrIS.close();
			} catch (IOException e) {
				Message.printError(ExceptionUtils.getStackTrace(e));
			}
		}
	}
	
	/**
	 * 
	 * 取得和 Search 的所有互動字串。
	 * 
	 * @return 和 Search 的所有互動字串。
	 */
	protected String getInteractiveLog(){
		return commandLogSB.toString();
	}
}
